#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.eventbus
      @file: eventbus.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
import inspect
from functools import partial

from hspylib.core.exception.exceptions import HSBaseException
from hspylib.core.preconditions import check_argument
from hspylib.modules.eventbus.event import Event
from typing import Any, Callable, Dict, List

EVENT_CALLBACK = Callable[[Event], None]


def subscribe(bus: str, events: str | list[str]):
    """Decorator to subscribe to a given bus event."""

    def helper(fn: Callable):
        """'subscribe' wrapper to handle both instance methods and functions."""

        handler_signature = inspect.signature(fn)
        args: list = []

        for param in handler_signature.parameters.values():
            if param.default == inspect.Parameter.empty and param.kind not in \
                [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]:
                args.append(param)
        count = len(args)

        check_argument(
            count >= 1, f"Event callbacks require the following arguments: (event). Given: '{count}'")

        def wrapped_function():
            if len(args) > 1 and inspect.isclass(type(args[0])):
                self = args[0]  # The first argument is 'self'
                return EventBus.get(bus).subscribe(events, partial(fn, args[0]))
            return EventBus.get(bus).subscribe(events, fn)

        return wrapped_function()

    return helper


def emit(bus_name: str, event_name: str, **kwargs):
    """Decorator to emit an event to the specified bus.
    :param bus_name: The name of the event bus.
    :param event_name: The name of the event.
    """
    EventBus.get(bus_name).emit(event_name, **kwargs)


class EventBus:
    """Provide an eventbus pattern for events and subscribers."""

    _buses: Dict[str, Any] = {}
    _subscribers: Dict[str, Any] = {}
    _events: List[Event] = []

    @classmethod
    def get(cls, bus_name: str) -> "EventBus":
        """Return the bus instance referred to the specified bus name.
        :param bus_name: The name of the event bus.
        """
        if bus_name in cls._buses:
            return cls._buses[bus_name]
        bus_instance = EventBus(bus_name)
        cls._buses[bus_name] = bus_instance
        return bus_instance

    @classmethod
    def _get_subscriber(cls, bus_name: str, event_name: str) -> Any:
        """Return the subscriber of the referred bus name and event name.
        :param bus_name: The name of the event bus.
        :param event_name: The name of the event.
        """
        cache_key = f"{bus_name}.{event_name}"
        if cache_key in cls._subscribers:
            return cls._subscribers[cache_key]
        subscriber = {"callbacks": []}
        cls._subscribers[cache_key] = subscriber
        return subscriber

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def subscribe(self, events: str | list[str], cb_event_handler: EVENT_CALLBACK) -> None:
        """Subscribe to the specified event bus.
        :param events: The name of the events.
        :param cb_event_handler: A callback that handles the event.
        """
        events = [events] if isinstance(events, str) else events
        for ev in events:
            subscriber = self._get_subscriber(self.name, ev)
            subscriber["callbacks"].append(cb_event_handler)

    def emit(self, event_name: str, **kwargs) -> None:
        """Emit an event to this bus.
        :param event_name: The name of the event.
        :param kwargs: The event keyword arguments.
        """
        self._events.append(Event(event_name, **kwargs))
        while len(self._events) > 0:
            event = self._events.pop()
            cache_key = f"{self.name}.{event.name}"
            subscribers = self._subscribers[cache_key] if cache_key in self._subscribers else None
            if subscribers and len(subscribers["callbacks"]) > 0:
                for callback in subscribers["callbacks"]:
                    try:
                        callback(event)
                    except Exception as err:
                        raise HSBaseException(
                            f"{self.__class__.__name__}::emit Callback invocation failed - {str(err)}"
                        ) from err



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.config
      @file: app_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
import logging as log
import os
import re
from textwrap import dedent
from typing import Any, Optional, TypeAlias

from hspylib.core.config.path_object import PathObject
from hspylib.core.config.properties import Properties
from hspylib.core.metaclass.classpath import AnyPath
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import root_dir, to_bool
from hspylib.core.tools.dict_tools import get_or_default_by_key

Placeholder: TypeAlias = None | dict[str, Any]


class AppConfigs(metaclass=Singleton):
    """Holds all of the application configurations (properties)"""

    INSTANCE = None

    DISPLAY_FORMAT = dedent(
        """
    AppConfigs
      |-Run-dir = {}
      |-Resources-dir = {}
      |-Properties:
       \\-{}
    """
    )

    @staticmethod
    def _replace_holders(value: str, placeholders: Placeholder) -> str:
        """Replace all placeholders by their associated value."""
        if value:
            replacements = {k.casefold(): v for k, v in placeholders.items()} if placeholders else {}
            re_ph = r'\$\{(\w+)\}'
            pattern = re.compile(re_ph, flags=re.IGNORECASE)
            while match := pattern.search(value):
                key = match.group(1)
                repl = os.environ.get(key, str(get_or_default_by_key(replacements, key, None)))
                re_placeholder = r'\$\{' + re.escape(key) + r'\}'
                value = re.sub(re_placeholder, repl, value, flags=re.IGNORECASE)

        return value

    def __init__(self, resource_dir: AnyPath, filename: str | None = None, profile: str | None = None):
        path_obj = PathObject.of(resource_dir)
        check_argument(path_obj.exists, "Unable to locate resources dir: {}", resource_dir)
        self._resource_dir = str(path_obj)
        self._properties = Properties(filename=filename, load_dir=resource_dir, profile=profile)
        log.info(self)

    # pylint: disable=consider-using-f-string
    def __str__(self) -> str:
        return "\n{}{}{}".format(
            "-=" * 40,
            self.DISPLAY_FORMAT.format(
                str(root_dir()),
                str(self._resource_dir),
                str(self._properties).replace(os.linesep, f"{os.linesep}   |-") if self._properties.size > 0 else "",
            ),
            "-=" * 40,
        )

    def __repr__(self):
        return str(self)

    def __getitem__(self, item: str) -> Optional[str]:
        return self.get(item)

    def __len__(self) -> int:
        return self.size

    @property
    def resource_dir(self) -> Optional[str]:
        """Return the configured application resource dir."""
        return self._resource_dir

    @property
    def properties(self) -> Properties:
        """Return the application properties."""
        return self._properties

    @property
    def size(self) -> int:
        """Return the application properties."""
        return self._properties.size

    def get(self, key: str, placeholders: Placeholder = None) -> Optional[Any]:
        """Get the value, as a string, of a property specified by key, otherwise None is returned.
        :param key: the key name of the property.
        :param placeholders: optional placeholders replacement.
        """
        return self._replace_holders(self._properties.get(key), placeholders)

    def get_int(self, key: str, placeholders: Placeholder = None) -> Optional[int]:
        """Get the value, as an integer, of a property specified by key, otherwise None is returned."""
        return Properties.convert_type(self._replace_holders(self._properties.get(key), placeholders), int)

    def get_float(self, key: str, placeholders: Placeholder = None) -> Optional[float]:
        """Get the value, as a float, of a property specified by key, otherwise None is returned."""
        return Properties.convert_type(self._replace_holders(self._properties.get(key), placeholders), float)

    def get_bool(self, key: str, placeholders: Placeholder = None) -> Optional[bool]:
        """Get the value, as a boolean, of a property specified by key, otherwise None is returned."""
        return Properties.convert_type(self._replace_holders(self._properties.get(key), placeholders), to_bool)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.decorator
      @file: decorators.py
   @created: Tue, 11 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""
import inspect
import logging as log
import os
import signal
import unittest
from datetime import datetime, timedelta
from functools import partial
from typing import Any, Callable

from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import to_bool


def integration_test(cls: type):
    """Enable/Disable HsPyLib integration tests"""

    it_disabled = to_bool(os.environ.get("HSPYLIB_IT_DISABLED", "True"))
    if it_disabled:
        log.debug("Skipping test: %s", cls.__name__)
        return unittest.skipIf(
            it_disabled, f"Disabled = {it_disabled} :integration tests because it needs real servers running"
        )

    return cls


def profiled(func: Callable):
    """Log the time spent on the Callable invocation."""

    def timedelta_to_str(delta: timedelta) -> str:
        """Convert timedelta object into string."""
        parts = str(delta).split(":")
        h, m, s = parts[0], parts[1], parts[2][:-7]
        ms = int(int(parts[2][-6:]) / 1000)
        return f"{h:02s}[h] {m:02s}[m] {s:02s}[s] {ms:02d}[ms]"

    def profiled_closure(*args, **kwargs) -> Any:
        """Execute the callable and return."""
        start = datetime.now()
        ret = func(*args, **kwargs)
        elapsed = datetime.now() - start
        print(f"@@@ [{func.__name__:>20}] Time elapsed\t{timedelta_to_str(elapsed)}")
        return ret

    return profiled_closure


def hook_signals(signals: tuple[signal.Signals, ...] | signal.Signals):
    """
    Decorator to hook a function to specific signals. When the specified signal(s) occur, the decorated function
    will be invoked.

    :param signals: A tuple of signal numbers or a single signal. These signals will be hooked to the function.
    """

    def helper(func: Callable):
        """Helper function that takes the function to be decorated, hooks it to the specified signal(s), and returns
        the wrapped function."""
        handler_signature = inspect.signature(func)
        args: list = []
        for param in handler_signature.parameters.values():
            if param.default == inspect.Parameter.empty and param.kind not in \
                [inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD]:
                args.append(param)
        count = len(args)

        check_argument(
            count >= 2, f"Signal handlers require the following arguments: (signum and frame). Given: '{count}'")

        sig_numbers = signals if isinstance(signals, tuple) else (signals,)

        def wrapped_function():
            """'helper' wrapper that supports instance methods as well as functions."""

            # If the first argument is 'self', bind the signal handler accordingly
            if count > 2 and inspect.isclass(type(args[0])):
                for s in sig_numbers:
                    signal.signal(s, partial(func, args[0]))
            else:
                for s in sig_numbers:
                    signal.signal(s, func)

        # Immediately set up the signal handler
        wrapped_function()

        return wrapped_function

    return helper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: fluid.py
   @created: Fri, 5 Jan 2024
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright (c) 2024, HomeSetup
"""

from functools import partial
from typing import Callable

from hspylib.core.namespace import Namespace
from hspylib.core.preconditions import check_not_none
from hspylib.modules.eventbus.eventbus import EventBus, emit


class FluidEvent:
    """Provide a 'Fluid' event base class."""

    def __init__(self, name: str, **kwargs):
        check_not_none(name, 'Parameter name is required!')
        self.name = name
        if 'cb_event_handler' not in kwargs:
            kwargs['cb_event_handler'] = None
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __str__(self):
        return f"FluidEvent-{self.name}::({', '.join(vars(self))})"

    def emit(self, bus_name: str, event_name: str, **kwargs) -> None:
        """Wrapper to the Event's emit method."""
        ...

    def subscribe(self, cb_event_handler: Callable) -> None:
        """Wrapper to the EventBus's subscribe method."""
        ...


class FluidEventBus:
    """Provide a 'Fluid' event bus base class."""

    def __init__(self, bus_name: str, **kwargs):
        self.name = bus_name
        self.bus = EventBus.get(bus_name)
        for key, evt in kwargs.items():
            fn_emit: Callable = partial(emit, bus_name, evt.name, **vars(evt))
            fn_subscribe: Callable = partial(self.bus.subscribe, evt.name, cb_event_handler=evt.cb_event_handler)
            setattr(evt, "emit", fn_emit)
            setattr(evt, "subscribe", fn_subscribe)
        self.events: Namespace = Namespace(f"FluidEventBus::{bus_name}", True, **kwargs)

    def __str__(self):
        return f"FluidEventBus-{self.bus.name}::({self.events})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.enums
      @file: charset.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright·(c)·2024,·HSPyLib
"""

from hspylib.core.enums.enumeration import Enumeration


class Charset(Enumeration):
    """Enumeration to wrap all standard charsets."""

    # fmt: off

    ASCII           = 'ascii'
    BIG5            = 'big5'
    BIG5HKSCS       = 'big5hkscs'

    CP037           = 'cp037'
    CP424           = 'cp424'
    CP437           = 'cp437'
    CP500           = 'cp500'
    CP737           = 'cp737'
    CP775           = 'cp775'
    CP850           = 'cp850'
    CP852           = 'cp852'
    CP855           = 'cp855'
    CP856           = 'cp856'
    CP857           = 'cp857'
    CP860           = 'cp860'
    CP861           = 'cp861'
    CP862           = 'cp862'
    CP863           = 'cp863'
    CP864           = 'cp864'
    CP865           = 'cp865'
    CP866           = 'cp866'
    CP869           = 'cp869'
    CP874           = 'cp874'
    CP875           = 'cp875'
    CP932           = 'cp932'
    CP949           = 'cp949'
    CP950           = 'cp950'
    CP1006          = 'cp1006'
    CP1026          = 'cp1026'
    CP1140          = 'cp1140'
    CP1250          = 'cp1250'
    CP1251          = 'cp1251'
    CP1252          = 'cp1252'
    CP1253          = 'cp1253'
    CP1254          = 'cp1254'
    CP1255          = 'cp1255'
    CP1256          = 'cp1256'
    CP1257          = 'cp1257'
    CP1258          = 'cp1258'

    EUC_JP          = 'euc-jp'
    EUC_JIS_2004    = 'euc-jis_2004'
    EUC_JISX0213    = 'euc-jisx0213'
    EUC_KR          = 'euc-kr'
    GB2312          = 'gb2312'
    GBK             = 'gbk'
    GB18030         = 'gb18030'
    HZ              = 'hz'

    ISO2022_JP      = 'iso2022-jp'
    ISO2022_JP_1    = 'iso2022-jp-1'
    ISO2022_JP_2    = 'iso2022-jp-2'
    ISO2022_JP_2004 = 'iso2022-jp-2004'
    ISO2022_JP_3    = 'iso2022-jp-3'
    ISO2022_JP_EXT  = 'iso2022-jp_ext'
    ISO2022_KR      = 'iso2022-kr'

    LATIN_1         = 'latin-1'
    ISO8859_1       = 'iso8859-1'
    ISO8859_2       = 'iso8859-2'
    ISO8859_3       = 'iso8859-3'
    ISO8859_4       = 'iso8859-4'
    ISO8859_5       = 'iso8859-5'
    ISO8859_6       = 'iso8859-6'
    ISO8859_7       = 'iso8859-7'
    ISO8859_8       = 'iso8859-8'
    ISO8859_9       = 'iso8859-9'
    ISO8859_10      = 'iso8859-10'
    ISO8859_13      = 'iso8859-13'
    ISO8859_14      = 'iso8859-14'
    ISO8859_15      = 'iso8859-15'

    JOHAB           = 'johab'
    KOI8_R          = 'koi8-r'
    KOI8_U          = 'koi8-u'

    MAC_CYRILLIC    = 'mac-cyrillic'
    MAC_GREEK       = 'mac-greek'
    MAC_ICELAND     = 'mac-iceland'
    MAC_LATIN2      = 'mac-latin2'
    MAC_ROMAN       = 'mac-roman'
    MAC_TURKISH     = 'mac-turkish'

    PTCP154         = 'ptcp154'
    SHIFT_JIS       = 'shift-jis'
    SHIFT_JIS_2004  = 'shift-jis_2004'
    SHIFT_JISX0213  = 'shift-jisx0213'

    UTF_16          = 'utf-16'
    UTF_16_BE       = 'utf-16-be'
    UTF_16_LE       = 'utf-16-le'
    UTF_7           = 'utf-7'
    UTF_8           = 'utf-8'

    # fmt: on

    @property
    def val(self) -> str:
        return str(self.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: cron_utils.py
   @created: Fri, 28 Feb 2024
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright (c) 2024, HomeSetup
"""

from datetime import datetime
from hspylib.core.metaclass.singleton import Singleton
from textwrap import dedent

import re


class CronUtils(metaclass=Singleton):
    """Provide Crontab utilities."""

    INSTANCE: 'CronUtils'

    # Cron based days of the week.
    DAY_NAME: list[str] = [
        'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'
    ]

    # Calendar month names.
    MONTH_NAME: list[str] = [
        'January', 'February', 'March', 'April', 'May', 'June',
        'July', 'August', 'September', 'October', 'November', 'December'
    ]

    # Define regex pattern for simple time; asterisks allowed.

    # Define regex pattern for ISO 8601 date-time; asterisks allowed.
    RE_ISO8601_DATETIME: str = r"^(\d{4}|\*)[-/](\d{2}|\*)[-/](\d{2}|\*)[T\s](\d{2}|\*):(\d{2}|\*):(\d{2}|\*)?$"

    @classmethod
    def ordinal(
        cls,
        str_num: str
    ) -> str:
        """Return the ordinal representation of a number."""
        if str_num.isdecimal():
            num = int(str_num)
            if 11 <= (num % 100) <= 13:
                suffix = 'th'
            else:
                suffix = ['th', 'st', 'nd', 'rd', 'th'][min(num % 10, 4)]
            return str(num) + suffix
        return str_num

    @classmethod
    def iso_to_cron(
        cls,
        iso_string: str,
        every_minute: bool = False,
        every_hour: bool = False,
        every_day: bool = False,
        every_month: bool = False,
        every_weekday: bool = False
    ) -> str:
        """Convert an ISO 8601 formatted date-time string to a cron expression.
        :param iso_string: ISO 8601 formatted date-time string (e.g., "2024-07-08T14:*:00"). Can include asterisks "*".
        :param every_minute: If True ignore the minute value and set it in the cron expression as '*'.
        :param every_hour: If True ignore the hour value and set it in the cron expression as '*'.
        :param every_day: If True ignore the day value and set it in the cron expression as '*'.
        :param every_month: If True ignore the month value and set it in the cron expression as '*'.
        :param every_weekday: If True ignore the weekday value and set it in the cron expression as '*'.
        :return str: A cron expression string (e.g., "45 14 8 7 1").
        """
        iso_pattern = re.compile(cls.RE_ISO8601_DATETIME)

        # Validate the ISO formatted date-time string
        if not iso_pattern.match(iso_string):
            raise ValueError("Invalid ISO date-time format")

        # Parse the ISO formatted date-time string
        parts = re.split('[T ]', iso_string)
        date_part = parts[0] if len(parts) > 0 else ''
        time_part = parts[1] if len(parts) > 1 else parts[0]

        c_date = re.split('[-/]', date_part)
        c_time = time_part.split(':')

        year, month, day = c_date[0], c_date[1], c_date[2]
        hour, minute = c_time[0], c_time[1]

        # Determine the day of the week if not using '*'
        if year != '*' and month != '*' and day != '*':
            dt = datetime(int(year), int(month), int(day))
            weekday = (dt.weekday() + 1) % 7  # Convert to cron format (Sunday is '0' or '7')
        else:
            weekday = '*'

        # Adjust fields based on boolean parameters
        minute = '*' if every_minute else f"{int(minute) if minute != '*' else '*'}"
        hour = '*' if every_hour else f"{int(hour) if hour != '*' else '*'}"
        day = '*' if every_day else f"{int(day) if day != '*' else '*'}"
        month = '*' if every_month else f"{int(month) if month != '*' else '*'}"
        weekday = '*' if every_weekday else f"{int(weekday) if weekday != '*' else '*'}"

        return f"{minute} {hour} {day} {month} {weekday}"

    @classmethod
    def human_readable_cron(
        cls,
        cron_expression: str
    ) -> str:
        """
        Generate a human-readable explanation for the given cron expression.

        Parameters:
        cron_expression (str): A cron expression string (e.g., "45 14 8 7 1").

        Returns:
        str: A human-readable explanation of the cron expression.
        """
        components = cron_expression.split()

        if len(components) != 5:
            raise ValueError("Invalid cron expression. It should have exactly 5 components.")

        minute, hour, day, month, weekday = components
        explanation = []
        dom = cls.ordinal(day)

        explanation.append('Runs')
        if hour == '*' and day == '*' and month == '*' and weekday == '*':
            explanation.append(f"every {minute + ' minute(s)' if minute != '*' else 'minute'} every day.")
        else:
            if hour == '*':
                explanation.append(f"every {cls.ordinal(minute) + ' ' if minute != '*' else ''}minute of every hour")
            else:
                time = f"at {minute if minute != '*' else 'every minute'}"
                time += f" past {hour}"
                explanation.append(f"{time}")

            if day != '*' and month != '*' and weekday != '*':
                explanation.append(
                    f"on the {dom} and every {cls.DAY_NAME[int(weekday)]} in {cls.MONTH_NAME[int(month) - 1]}.")
            elif day != '*' and month != '*' and weekday == '*':
                explanation.append(f"on {cls.MONTH_NAME[int(month) - 1]} {dom}.")
            elif day == '*' and month != '*' and weekday != '*':
                explanation.append(f"every {cls.DAY_NAME[int(weekday)]} in {cls.MONTH_NAME[int(month) - 1]}.")
            elif day != '*' and month == '*' and weekday != '*':
                explanation.append(f"every month on {cls.DAY_NAME[int(weekday)]}.")
            elif day == '*' and month == '*' and weekday != '*':
                explanation.append(f"every {cls.DAY_NAME[int(weekday)]}.")
            elif day != '*' and month == '*' and weekday == '*':
                explanation.append(f"every month, on the {dom} day.")
            elif day == '*' and month != '*' and weekday == '*':
                explanation.append(f"every day in {cls.MONTH_NAME[int(month) - 1]}.")
            elif day != '*' and month == '*' and weekday == '*':
                explanation.append(f"every month, on the {dom} day.")
            else:
                explanation.append(f", every day.")

        return ' '.join(explanation).replace(" .", ".").replace("the *th ", "every ").replace('* minutes', 'minute')

    @classmethod
    def explain_cron(
        cls,
        cron_exp: str
    ) -> str:
        """Explain a cron expression in human-readable format.
        :param cron_exp: A cron expression string (e.g., "45 14 * * 1").
        :return str: A detailed explanation of the cron expression.
        """

        def sep(digits: str | int) -> str:
            sep_str: str = ' ' * len(str(digits))
            return f'|{sep_str}'

        if (components := cron_exp.split()) and (lc := len(components)) != 5:
            raise ValueError(f"Invalid cron expression. It should have exactly 5 components, but found {lc}")

        # Split the cron expression into its components
        minute, hour, day, month, weekday = components
        padding = 2
        s4 = f"{sep(minute)}{sep(hour)}{sep(day)}{sep(month)}"
        s3 = f"{sep(minute)}{sep(hour)}{sep(day)}"
        s2 = f"{sep(minute)}{sep(hour)}"
        s1 = f"{sep(minute)}"

        l5 = len(cron_exp) + padding
        l4 = l5 - len(s1)
        l3 = l5 - len(s2)
        l2 = l5 - len(s3)
        l1 = l5 - len(s4)

        d5 = f"+{'-' * l5}"
        d4 = f"+{'-' * l4}"
        d3 = f"+{'-' * l3}"
        d2 = f"+{'-' * l2}"
        d1 = f"+{'-' * l1}"

        # Create explanations for each component
        moy_str: str = cls.MONTH_NAME[int(month) - 1] if month != '*' else '*'
        dow_str: str = cls.DAY_NAME[int(weekday)] if weekday != '*' else '*'

        minute_exp = f"{d5} [0-59]  Minute: '{cls.ordinal(minute) + ' minute' if minute != '*' else 'Every minute'}'"
        hour_exp = f"{s1}{d4} [0-23]    Hour: '{cls.ordinal(hour) + ' hour' if hour != '*' else 'Every hour'}'"
        day_exp = f"{s2}{d3} [1-31]     Day: '{cls.ordinal(day) + ' day' if day != '*' else 'Every day'}'"
        month_exp = f"{s3}{d2} [1-12]   Month: '{moy_str if month != '*' else 'Every month'}'"
        day_of_week_exp = f"{s4}{d1} [0-7]  Weekday: '{dow_str if dow_str != '*' else 'Every weekday'}'"

        # Combine explanations into a single string
        explanation = dedent(f"""
        {cron_exp}
        {'-' * len(minute)} {'-' * len(hour)} {'-' * len(day)} {'-' * len(month)} {'-' * len(weekday)}
        {sep(minute)}{sep(hour)}{sep(day)}{sep(month)}{sep(weekday)}
        {day_of_week_exp}
        {month_exp}
        {day_exp}
        {hour_exp}
        {minute_exp}
        """).strip()

        return f"\n{explanation}\n"


assert (cron := CronUtils().INSTANCE) is not None

